from datetime import datetime
from dataclasses import dataclass
from typing import List
import logging
import time
import requests
import threading
import os
import pickle

@dataclass()
class ConnectionInfo:
    date: datetime
    is_connected: bool
    ip_address: str
    isp: str


class ConnectionLog:

    def __init__(self, filename:str = None):
        if filename is None:
            dir = os.path.join("var", "lib", "netmonitor")
            os.makedirs(dir, exist_ok=True)
            self.filename = os.path.join(dir, "history.p")
        else:
            self.filename = filename

        try:
            with open(self.filename, "rb") as file:
                self.entries = pickle.load(file)
        except Exception as e:
            logging.error(e)
            self.entries = list()

    def append(self, connection_info : ConnectionInfo):
        if len(self.entries) > 500:
            del self.entries[0]
        self.entries.append(connection_info)
        self.__store()

    def __store(self):
        try:
            with open(self.filename, "wb") as file:
                pickle.dump(self.entries, file)
        except Exception as e:
            logging.error(e)

    def print_duration(self, duration: int):
        if duration > (60 * 60):
            return "{0:.1f} h".format(duration/(60*60))
        elif duration > 60:
            return "{0:.1f} min".format(duration/60)
        else:
            return "{0:.1f} sec".format(duration)

    def to_report(self) -> List[str]:
        report = list()

        previous_entry = None
        for entry in self.entries:
            try:
                status = "connected" if entry.is_connected else "disconnected"
                detail = ""
                if previous_entry is not None:
                    elapsed_sec = int((entry.date - previous_entry.date).total_seconds())
                    if entry.is_connected and not previous_entry.is_connected:
                        detail = "reconnected after " + self.print_duration(elapsed_sec)
                    elif len(entry.ip_address) > 0 and len(previous_entry.ip_address) > 0 and entry.ip_address != previous_entry.ip_address:
                        detail = "ip address updated"
                report.append(entry.date.strftime("%Y-%m-%d %H:%M:%S") + ", " + status + ", " + entry.ip_address + ", " + entry.isp + ", " + detail)
                previous_entry = entry
            except Exception as e:
                print(e)
        return report


class IpAddressResolver:

    def __init__(self):
        self.cache_ip_address = ""
        self.cache_time = datetime.fromtimestamp(555)

    def get_internet_address(self, max_cache_ttl: int = 0) -> str:
        try:
            now = datetime.now()
            if max_cache_ttl == 0 or (now - self.cache_time).seconds > max_cache_ttl:
                response = requests.get('http://whatismyip.akamai.com/', timeout=10)
                if (response.status_code >= 200) and (response.status_code < 300):
                    self.cache_ip_address = response.text
                    self.cache_time = now
                    logging.info('ip address resolved ' + self.cache_ip_address)
            return self.cache_ip_address
        except Exception as e:
            return ""


class IpInfo:

    def __init__(self):
        self.cache= dict()
        self.cached_invalidation_time = datetime.fromtimestamp(555)

    def get_ip_info(self, ip: str) -> str:
        try:
            if (datetime.now() - self.cached_invalidation_time).seconds > (4 * 24 * 60 * 60):
                self.cache = dict()
                self.cached_invalidation_time = datetime.now()
                logging.info('ip info cache invalidated')
            if ip not in self.cache.keys():
                response = requests.get('https://tools.keycdn.com/geo.json?host=' + ip, timeout=60)
                if (response.status_code >= 200) and (response.status_code < 300):
                    data = response.json()
                    self.cache[ip] = data['data']['geo']['isp']
                    logging.info('ip info fetched ' + ip + ":" + self.cache[ip])
            return self.cache.get(ip, "")
        except Exception as e:
            return ""


class ConnectionTester:

    def __init__(self, connection_log : ConnectionLog):
        self.connection_log = connection_log
        self.address_resolver = IpAddressResolver()
        self.ip_info = IpInfo()

    def listen(self, listener, measure_period_sec, test_uri: str = "http://google.com"):
        threading.Thread(target=self.__measure_periodically, args=(measure_period_sec, test_uri, listener), daemon=True).start()

    def measure(self, test_uri, max_cache_ttl: int) -> ConnectionInfo:
        try:
            requests.get(test_uri, timeout=7) # test call
            ip_address = self.address_resolver.get_internet_address(max_cache_ttl)
            isp = self.ip_info.get_ip_info(ip_address)
            return ConnectionInfo(datetime.now(), True, ip_address, isp)
        except:
            return ConnectionInfo(datetime.now(), False, "", "")

    def __measure_periodically(self, measure_period_sec: int, test_uri: str, listener):
        previous_connection_info = None
        while True:
            try:
                if previous_connection_info is None or not previous_connection_info.is_connected:
                    connected_info = self.measure(test_uri, max_cache_ttl=0)
                else:
                    connected_info = self.measure(test_uri, max_cache_ttl=112)
                if previous_connection_info is None or self.is_connect_state_changed(connected_info, previous_connection_info) or self.is_ip_address_state_changed(connected_info, previous_connection_info):
                    self.connection_log.append(connected_info)
                    listener(connected_info)
                    previous_connection_info = connected_info
            except Exception as e:
                logging.error(e)
            time.sleep(measure_period_sec)

    def is_connect_state_changed(self, current: ConnectionInfo, previous: ConnectionInfo):
       return current.is_connected != previous.is_connected

    def is_ip_address_state_changed(self, current: ConnectionInfo, previous: ConnectionInfo):
        return current.ip_address != previous.ip_address


from speedtest import Speedtest
from dataclasses import dataclass
import threading
import time
import logging


@dataclass
class Speed:
    server: str
    downloadspeed: int
    uploadspeed: int
    ping: float
    report_uri: str


class SpeedtestRunner:

    def __init__(self, listener):
        self.listener = listener

    def run_periodically(self, measure_period_sec: int):
        threading.Thread(target=self.__measure_periodically, args=(measure_period_sec,), daemon=True).start()

    def __measure_periodically(self, measure_period_sec: int):
        while True:
            try:
                speed = self.measure()
                self.listener(speed)
            except Exception as e:
                logging.error(e)
            time.sleep(measure_period_sec)

    def measure(self) -> Speed:
        s = Speedtest()
        s.download()
        s.upload()
        try:
            link = s.results.share()   # POST data to the speedtest.net API to obtain a share results link
        except:
            link = None
        metrics = s.results.dict()
        return Speed(metrics['server'].get('sponsor', '') + "/" + metrics['server'].get('name', ''), int(metrics['download']), int(metrics['upload']), metrics['ping'], link)



from webthing import Property, Thing, Value
from internet_monitor_webthing.connectivity_monitor import ConnectionInfo, ConnectionLog, ConnectionTester
import tornado.ioloop


class InternetConnectivityMonitorWebthing(Thing):

    # regarding capabilities refer https://iot.mozilla.org/schemas
    # there is also another schema registry http://iotschema.org/docs/full.html not used by webthing

    def __init__(self, description: str, connecttest_period: int, connecttest_url: str):
        Thing.__init__(
            self,
            'urn:dev:ops:connectivitymonitor-1',
            'Internet Connectivity Monitor',
            ['MultiLevelSensor'],
            description
        )
        self.history = ConnectionLog()
        self.connecttest_period = connecttest_period

        self.internet_connected = Value(False)
        self.add_property(
            Property(self,
                     'connected',
                     self.internet_connected,
                     metadata={
                         '@type': 'BooleanProperty',
                         'title': 'Internet is connected',
                         "type": "boolean",
                         'description': 'Whether the internet is connected',
                         'readOnly': True,
                     }))

        self.test_url = Value(connecttest_url)
        self.add_property(
            Property(self,
                     'test_url',
                     self.test_url,
                     metadata={
                         '@type': 'Name',
                         'title': 'Internet connection test url',
                         "type": "string",
                         'description': 'The url to connect',
                         'readOnly': True,
                     }))

        self.testperiod = Value(connecttest_period)
        self.add_property(
            Property(self,
                     'test_period',
                     self.testperiod,
                     metadata={
                         '@type': 'LevelProperty',
                         'title': 'Internet connection test execution period in seconds',
                         'type': 'number',
                         'description': 'The Internet connection test execution period in seconds',
                         'unit': 'sec',
                         'readOnly': True,
                     }))

        self.ip_address = Value("")
        self.add_property(
            Property(self,
                     'ip_address',
                     self.ip_address,
                     metadata={
                         'title': 'Public IP address',
                         'type': 'string',
                         'description': 'The public WAN IP address used for internet connection',
                         'readOnly': True,
                     }))

        self.isp = Value("")
        self.add_property(
            Property(self,
                     'isp',
                     self.isp,
                     metadata={
                         'title': 'Internet service provider',
                         'type': 'string',
                         'description': 'The name of the internet service provider providing the public WAN IP address ',
                         'readOnly': True,
                     }))

        self.connection_history = Value("")
        self.add_property(
            Property(self,
                     'connection_history',
                     self.connection_history,
                     metadata={
                         'title': 'Availability report',
                         'type': 'array',
                         'description': 'The availability report',
                         'readOnly': True,
                     }))

        self.ioloop = tornado.ioloop.IOLoop.current()
        self.tester = ConnectionTester(self.history)
        self.tester.listen(self.__connection_state_updated, self.testperiod.get(), self.test_url.get())

    def __connection_state_updated(self, connection_info: ConnectionInfo):
        self.ioloop.add_callback(self.__update_connected_props, connection_info)

    def __update_connected_props(self, connection_info: ConnectionInfo):
        self.internet_connected.notify_of_external_update(connection_info.is_connected)
        self.ip_address.notify_of_external_update(connection_info.ip_address)
        self.isp.notify_of_external_update(connection_info.isp)
        self.connection_history.notify_of_external_update(self.history.to_report())



from webthing import (Property, Thing, Value, Action)
from internet_monitor_webthing.speedtest_monitor import SpeedtestRunner, Speed
from datetime import datetime
import tornado.ioloop
import uuid


class TriggerSpeedTest(Action):

    def __init__(self, thing, input_):
        Action.__init__(self, uuid.uuid4().hex, thing, 'trigger', input_=input_)

    def perform_action(self):
        self.thing.speedtest_runner.measure()


class InternetSpeedMonitorWebthing(Thing):

    # regarding capabilities refer https://iot.mozilla.org/schemas
    # there is also another schema registry http://iotschema.org/docs/full.html not used by webthing

    def __init__(self, description: str, speedtest_period: int):
        Thing.__init__(
            self,
            'urn:dev:ops:speedmonitor-1',
            'Internet Speed Monitor',
            ['MultiLevelSensor'],
            description
        )

        self.downloadspeed = Value(0)
        self.add_property(
            Property(self,
                     'download_speed',
                     self.downloadspeed,
                     metadata={
                         '@type': 'LevelProperty',
                         'title': 'Internet download speed',
                         'type': 'number',
                         'description': 'The current internet download  speed',
                         'unit': 'Mbit/sec',
                         'readOnly': True,
                     }))

        self.uploadspeed = Value(0)
        self.add_property(
            Property(self,
                     'upload_speed',
                     self.uploadspeed,
                     metadata={
                         '@type': 'LevelProperty',
                         'title': 'Internet upload speed',
                         'type': 'number',
                         'description': 'The current internet upload speed',
                         'unit': 'Mbit/sec',
                         'readOnly': True,
                     }))

        self.ping_time = Value(0)
        self.add_property(
            Property(self,
                     'ping',
                     self.ping_time,
                     metadata={
                         '@type': 'LevelProperty',
                         'title': 'Internet ping (latency)',
                         'type': 'number',
                         'description': 'The current internet ping latency',
                         'unit': 'milliseconds',
                         'readOnly': True,
                     }))

        self.testdate = Value("")
        self.add_property(
            Property(self,
                     'last_test',
                     self.testdate,
                     metadata={
                         'title': 'Last executed test date',
                         'type': 'string',
                         'description': 'The date of the last successfully executed test (iso 8601 string)',
                         'readOnly': True,
                     }))

        self.testperiod = Value(speedtest_period)
        self.add_property(
            Property(self,
                     'speedtest_exection_period',
                     self.testperiod,
                     metadata={
                         '@type': 'LevelProperty',
                         'title': 'Speedtest execution period',
                         'type': 'number',
                         'description': 'Speedtest execution period in seconds',
                         'unit': 'sec',
                         'readOnly': True,
                     }))

        self.testserver = Value("")
        self.add_property(
            Property(self,
                     'speedtest_server',
                     self.testserver,
                     metadata={
                         'title': 'Speedtest servername',
                         'type': 'string',
                         'description': 'The speedtest server which has been connected to perform the speedtest',
                         'readOnly': True,
                     }))

        self.resulturi = Value("")
        self.add_property(
            Property(self,
                     'speedtest_result_uri',
                     self.resulturi,
                     metadata={
                         'title': 'Speedtest result report url',
                         'type': 'string',
                         'description': 'speedtest result report url',
                         'readOnly': True,
                     }))

        self.ioloop = tornado.ioloop.IOLoop.current()
        self.speedtest_runner = SpeedtestRunner(self.__on_speed_updated)
        self.add_available_action(
            'trigger',
            {
                'title': 'Trigger',
                'description': 'Triggers a speed test run',
            },
            TriggerSpeedTest)
        self.speedtest_runner.run_periodically(self.testperiod.get())

    def __on_speed_updated(self, speed: Speed):
        self.ioloop.add_callback(self.__update_speed_props, speed)

    def __update_speed_props(self, speed: Speed):
        self.uploadspeed.notify_of_external_update(self.__to_mbit(speed.uploadspeed))
        self.downloadspeed.notify_of_external_update(self.__to_mbit(speed.downloadspeed))
        self.ping_time.notify_of_external_update(speed.ping)
        self.testdate.notify_of_external_update(datetime.now().isoformat())
        self.testserver.notify_of_external_update(speed.server)
        self.resulturi.notify_of_external_update(speed.report_uri)

    def __to_mbit(self, bit_pre_sec: int):
        return round(bit_pre_sec / (1000 * 1000), 1)
